@@ -1,11 +1,11 @@
 /* -*-pgsql-c-*- */
 /*
- * $Header: /cvsroot/pgpool/pgpool-II/child.c,v 1.65.2.4 2011/08/30 11:31:14 kitagawa Exp $
+ * $Header$
  *
  * pgpool: a language independent connection pool server for PostgreSQL
  * written by Tatsuo Ishii
  *
- * Copyright (c) 2003-2011	PgPool Global Development Group
+ * Copyright (c) 2003-2017	PgPool Global Development Group
  *
  * Permission to use, copy, modify, and distribute this software and
  * its documentation for any purpose and without fee is hereby
@@ -56,6 +56,7 @@
 #include "pool_ip.h"
 #include "md5.h"
 #include "pool_stream.h"
+#include "pool_passwd.h"
 
 static POOL_CONNECTION *do_accept(int unix_fd, int inet_fd, struct timeval *timeout);
 static StartupPacket *read_startup_packet(POOL_CONNECTION *cp);
@@ -74,6 +75,8 @@ static void init_system_db_connection(void);
 static bool connect_using_existing_connection(POOL_CONNECTION *frontend,
 											  POOL_CONNECTION_POOL *backend,
 											  StartupPacket *sp);
+static void free_persisten_db_connection_memory(POOL_CONNECTION_POOL_SLOT *cp);
+
 /*
  * non 0 means SIGTERM(smart shutdown) or SIGINT(fast shutdown) has arrived
  */
@@ -82,6 +85,9 @@ volatile sig_atomic_t exit_request = 0;
 static int idle;		/* non 0 means this child is in idle state */
 static int accepted = 0;
 
+static int child_inet_fd = 0;
+static int child_unix_fd = 0;
+
 extern int myargc;
 extern char **myargv;
 
@@ -89,6 +95,9 @@ char remote_ps_data[NI_MAXHOST];		/* used for set_ps_display */
 
 volatile sig_atomic_t got_sighup = 0;
 
+char remote_host[NI_MAXHOST];	/* client host */
+char remote_port[NI_MAXSERV];	/* client port */
+
 /*
 * child main loop
 */
@@ -128,13 +137,25 @@ void do_child(int unix_fd, int inet_fd)
 		pool_set_nonblock(inet_fd);
 	}
 #endif
+	if (inet_fd)
+	{
+		child_inet_fd = inet_fd;
+	}
+	child_unix_fd = unix_fd;
+
+	/* Initialize my backend status */
+	pool_initialize_private_backend_status();
 
 	/* Initialize per process context */
 	pool_init_process_context();
 
 	/* initialize random seed */
 	gettimeofday(&now, &tz);
+#if defined(sun) || defined(__sun)
+	srand((unsigned int) now.tv_usec);
+#else
 	srandom((unsigned int) now.tv_usec);
+#endif
 
 	/* initialize system db connection */
 	init_system_db_connection();
@@ -145,6 +166,15 @@ void do_child(int unix_fd, int inet_fd)
 		child_exit(1);
 	}
 
+	/*
+	 * Open pool_passwd in child process.  This is necessary to avoid the
+	 * file descriptor race condition reported in [pgpool-general: 1141].
+	 */
+	if (strcmp("", pool_config->pool_passwd))
+	{
+		pool_reopen_passwd_file();
+	}
+
 	timeout.tv_sec = pool_config->child_life_time;
 	timeout.tv_usec = 0;
 
@@ -157,6 +187,17 @@ void do_child(int unix_fd, int inet_fd)
 		/* pgpool stop request already sent? */
 		check_stop_request();
 
+		/* Check if restart request is set because of failback event
+		 * happend.  If so, exit myself with exit code 1 to be
+		 * restarted by pgpool parent.
+		 */
+		if (pool_get_my_process_info()->need_to_restart)
+		{
+			pool_log("do_child: failback event found. restart myself.");
+			pool_get_my_process_info()->need_to_restart = 0;
+			child_exit(1);
+		}
+
 		accepted = 0;
 
 		/* perform accept() */
@@ -218,6 +259,7 @@ void do_child(int unix_fd, int inet_fd)
 		{
 			pool_debug("SSLRequest from client");
 			pool_ssl_negotiate_serverclient(frontend);
+			pool_free_startup_packet(sp);
 			goto retry_startup;
 		}
 
@@ -251,6 +293,18 @@ void do_child(int unix_fd, int inet_fd)
 		 * Authentication is also done in this step.
 		 */
 
+		/* Check if restart request is set because of failback event
+		 * happend.  If so, close idle connections to backend and make
+		 * a new copy of backend status.
+		 */
+		if (pool_get_my_process_info()->need_to_restart)
+		{
+			pool_log("do_child: failback event found. discard existing connections");
+			pool_get_my_process_info()->need_to_restart = 0;
+			close_idle_connection(0);
+			pool_initialize_private_backend_status();
+		}
+
 		/*
 		 * if there's no connection associated with user and database,
 		 * we need to connect to the backend and send the startup packet.
@@ -317,6 +371,12 @@ void do_child(int unix_fd, int inet_fd)
 		 */
 		pool_init_session_context(frontend, backend);
 
+		/*
+		 * Set protocol versions
+		 */
+		pool_set_major_version(sp->major);
+		pool_set_minor_version(sp->minor);
+
 		/* Mark this connection pool is conncted from frontend */
 		pool_coninfo_set_frontend_connected(pool_get_process_context()->proc_id, pool_pool_index());
 
@@ -333,12 +393,15 @@ void do_child(int unix_fd, int inet_fd)
 			{
 				/* client exits */
 				case POOL_END:
+				case POOL_END_WITH_FRONTEND_ERROR:
 					/*
 					 * do not cache connection if:
+					 * frontend abnormally exits or
 					 * pool_config->connection_cahe == 0 or
 					 * database name is template0, template1, postgres or regression
 					 */
-					if (pool_config->connection_cache == 0 ||
+					if (status == POOL_END_WITH_FRONTEND_ERROR ||
+						pool_config->connection_cache == 0 ||
 						!strcmp(sp->database, "template0") ||
 						!strcmp(sp->database, "template1") ||
 						!strcmp(sp->database, "postgres") ||
@@ -453,9 +516,6 @@ static POOL_CONNECTION *do_accept(int unix_fd, int inet_fd, struct timeval *time
 	struct timeval *timeoutval;
 	struct timeval tv1, tv2, tmback = {0, 0};
 
-	char remote_host[NI_MAXHOST];
-	char remote_port[NI_MAXSERV];
-
 	set_ps_display("wait for connection request", false);
 
 	/* Destroy session context for just in case... */
@@ -608,7 +668,11 @@ static POOL_CONNECTION *do_accept(int unix_fd, int inet_fd, struct timeval *time
 	{
 		pool_get_config(get_config_file_name(), RELOAD_CONFIG);
 		if (pool_config->enable_pool_hba)
+		{
 			load_hba(get_hba_file_name());
+			if (strcmp("", pool_config->pool_passwd))
+				pool_reopen_passwd_file();
+		}
 		if (pool_config->parallel_mode)
 			pool_memset_system_db_info(system_db_info->info);
 		got_sighup = 0;
@@ -762,7 +826,9 @@ static POOL_CONNECTION *do_accept(int unix_fd, int inet_fd, struct timeval *time
 }
 
 /*
-* read startup packet
+* Read startup packet
+*
+* Read the startup packet and parse the contents.
 */
 static StartupPacket *read_startup_packet(POOL_CONNECTION *cp)
 {
@@ -781,26 +847,25 @@ static StartupPacket *read_startup_packet(POOL_CONNECTION *cp)
 
 	if (pool_config->authentication_timeout > 0)
 	{
-		pool_signal(SIGALRM, authentication_timeout);
-		alarm(pool_config->authentication_timeout);
+		pool_alarm(authentication_timeout, pool_config->authentication_timeout);
 	}
 
 	/* read startup packet length */
 	if (pool_read(cp, &len, sizeof(len)))
 	{
+		pool_error("read_startup_packet: incorrect packet length (%d)", len);
+		pool_free_startup_packet(sp);
+		pool_undo_alarm();
 		return NULL;
 	}
 	len = ntohl(len);
 	len -= sizeof(len);
 
-	if (len <= 0)
+	if (len <= 0 || len >= MAX_STARTUP_PACKET_LENGTH)
 	{
 		pool_error("read_startup_packet: incorrect packet length (%d)", len);
-	}
-	else if (len >= MAX_STARTUP_PACKET_LENGTH)
-	{
-		pool_error("read_startup_packet: invalid startup packet");
 		pool_free_startup_packet(sp);
+		pool_undo_alarm();
 		return NULL;
 	}
 
@@ -809,8 +874,7 @@ static StartupPacket *read_startup_packet(POOL_CONNECTION *cp)
 	{
 		pool_error("read_startup_packet: out of memory");
 		pool_free_startup_packet(sp);
-		alarm(0);
-		pool_signal(SIGALRM, SIG_IGN);
+		pool_undo_alarm();
 		return NULL;
 	}
 
@@ -818,8 +882,7 @@ static StartupPacket *read_startup_packet(POOL_CONNECTION *cp)
 	if (pool_read(cp, sp->startup_packet, len))
 	{
 		pool_free_startup_packet(sp);
-		alarm(0);
-		pool_signal(SIGALRM, SIG_IGN);
+		pool_undo_alarm();
 		return NULL;
 	}
 
@@ -839,8 +902,7 @@ static StartupPacket *read_startup_packet(POOL_CONNECTION *cp)
 			{
 				pool_error("read_startup_packet: out of memory");
 				pool_free_startup_packet(sp);
-				alarm(0);
-				pool_signal(SIGALRM, SIG_IGN);
+				pool_undo_alarm();
 				return NULL;
 			}
 			strncpy(sp->database, sp2->database, SM_DATABASE);
@@ -850,8 +912,7 @@ static StartupPacket *read_startup_packet(POOL_CONNECTION *cp)
 			{
 				pool_error("read_startup_packet: out of memory");
 				pool_free_startup_packet(sp);
-				alarm(0);
-				pool_signal(SIGALRM, SIG_IGN);
+				pool_undo_alarm();
 				return NULL;
 			}
 			strncpy(sp->user, sp2->user, SM_USER);
@@ -871,8 +932,7 @@ static StartupPacket *read_startup_packet(POOL_CONNECTION *cp)
 					{
 						pool_error("read_startup_packet: out of memory");
 						pool_free_startup_packet(sp);
-						alarm(0);
-						pool_signal(SIGALRM, SIG_IGN);
+						pool_undo_alarm();
 						return NULL;
 					}
 				}
@@ -884,11 +944,28 @@ static StartupPacket *read_startup_packet(POOL_CONNECTION *cp)
 					{
 						pool_error("read_startup_packet: out of memory");
 						pool_free_startup_packet(sp);
-						alarm(0);
-						pool_signal(SIGALRM, SIG_IGN);
+						pool_undo_alarm();
 						return NULL;
 					}
 				}
+
+				/*
+				 * From 9.0, the start up packet may include
+				 * application name. After receiving such that packet,
+				 * backend sends parameter status of application_name.
+				 * Upon reusing connection to backend, we need to
+				 * emulate this behavior of backend. So we remember
+				 * this and send parameter status packet to frontend
+				 * instead of backend in
+				 * connect_using_existing_connection().
+				 */
+				else if (!strcmp("application_name", p))
+				{
+					p += (strlen(p) + 1);
+					sp->application_name = p;
+					pool_debug("read_startup_packet: application_name: %s", p);
+				}
+
 				p += (strlen(p) + 1);
 			}
 			break;
@@ -900,8 +977,7 @@ static StartupPacket *read_startup_packet(POOL_CONNECTION *cp)
 			{
 				pool_error("read_startup_packet: out of memory");
 				pool_free_startup_packet(sp);
-				alarm(0);
-				pool_signal(SIGALRM, SIG_IGN);
+				pool_undo_alarm();
 				return NULL;
 			}
 			sp->user = calloc(1, 1);
@@ -909,8 +985,7 @@ static StartupPacket *read_startup_packet(POOL_CONNECTION *cp)
 			{
 				pool_error("read_startup_packet: out of memory");
 				pool_free_startup_packet(sp);
-				alarm(0);
-				pool_signal(SIGALRM, SIG_IGN);
+				pool_undo_alarm();
 				return NULL;
 			}
 			break;
@@ -918,21 +993,40 @@ static StartupPacket *read_startup_packet(POOL_CONNECTION *cp)
 		default:
 			pool_error("read_startup_packet: invalid major no: %d", sp->major);
 			pool_free_startup_packet(sp);
-			alarm(0);
-			pool_signal(SIGALRM, SIG_IGN);
+			pool_undo_alarm();
 			return NULL;
 	}
 
+	/* Check a user name was given. */
+	if (sp->major != 1234 &&
+	    (sp->user == NULL || sp->user[0] == '\0'))
+	{
+		pool_send_fatal_message(cp, sp->major, "28000",
+		                        "no PostgreSQL user name specified in startup packet",
+								"",
+								"",
+								__FILE__, __LINE__);
+		pool_error("read_startup_packet: no PostgreSQL user name specified in startup packet");
+		pool_free_startup_packet(sp);
+		pool_undo_alarm();
+		return NULL;
+	}
+
+	/* The database defaults to ther user name. */
+	if (sp->database == NULL || sp->database[0] == '\0')
+	{
+		sp->database = strdup(sp->user);
+	}
+
 	pool_debug("Protocol Major: %d Minor: %d database: %s user: %s",
 			   sp->major, sp->minor, sp->database, sp->user);
-	alarm(0);
-	pool_signal(SIGALRM, SIG_IGN);
+	pool_undo_alarm();
 	return sp;
 }
 
 /*
-* send startup packet
-*/
+ * send startup packet
+ */
 int send_startup_packet(POOL_CONNECTION_POOL_SLOT *cp)
 {
 	int len;
@@ -945,13 +1039,13 @@ int send_startup_packet(POOL_CONNECTION_POOL_SLOT *cp)
 /*
  * Reuse existing connection
  */
-static bool connect_using_existing_connection(POOL_CONNECTION *frontend, 
+static bool connect_using_existing_connection(POOL_CONNECTION *frontend,
 											  POOL_CONNECTION_POOL *backend,
 											  StartupPacket *sp)
 {
 	int i, freed = 0;
 	/*
-	 * save startup packet info
+	 * Save startup packet info
 	 */
 	for (i = 0; i < NUM_BACKENDS; i++)
 	{
@@ -966,7 +1060,7 @@ static bool connect_using_existing_connection(POOL_CONNECTION *frontend,
 		}
 	}
 
-	/* reuse existing connection to backend */
+	/* Reuse existing connection to backend */
 
 	if (pool_do_reauth(frontend, backend))
 	{
@@ -977,6 +1071,32 @@ static bool connect_using_existing_connection(POOL_CONNECTION *frontend,
 
 	if (MAJOR(backend) == 3)
 	{
+		char command_buf[1024];
+
+		/* If we have received application_name in the start up
+		 * packet, we send SET command to backend. Also we add or
+		 * replace existing application_name data.
+		 */
+		if (sp->application_name)
+		{
+			snprintf(command_buf, sizeof(command_buf), "SET application_name TO '%s'", sp->application_name);
+
+			for (i=0;i<NUM_BACKENDS;i++)
+			{
+				if (VALID_BACKEND(i))
+					if (do_command(frontend, CONNECTION(backend, i),
+							   command_buf, MAJOR(backend),
+								   MASTER_CONNECTION(backend)->pid,
+								   MASTER_CONNECTION(backend)->key, 0) != POOL_CONTINUE)
+					{
+						pool_error("connect_using_existing_connection: do_command failed. command: %s", command_buf);
+						return false;
+					}
+			}
+
+			pool_add_param(&MASTER(backend)->params, "application_name", sp->application_name);
+		}
+
 		if (send_params(frontend, backend))
 		{
 			pool_close(frontend);
@@ -985,7 +1105,7 @@ static bool connect_using_existing_connection(POOL_CONNECTION *frontend,
 		}
 	}
 
-	/* send ReadyForQuery to frontend */
+	/* Send ReadyForQuery to frontend */
 	pool_write(frontend, "Z", 1);
 
 	if (MAJOR(backend) == 3)
@@ -1057,7 +1177,7 @@ void cancel_request(CancelPacket *sp)
 		if (!VALID_BACKEND(i))
 			continue;
 
-		if (*(BACKEND_INFO(i).backend_hostname) == '\0')
+		if (*(BACKEND_INFO(i).backend_hostname) == '/')
 			fd = connect_unix_domain_socket(i, TRUE);
 		else
 			fd = connect_inet_domain_socket(i, TRUE);
@@ -1162,10 +1282,22 @@ static RETSIGTYPE die(int sig)
 	switch (sig)
 	{
 		case SIGTERM:	/* smart shutdown */
+			/* Refuse further requests by closing listen socket */
+			if (child_inet_fd)
+			{
+#ifdef NOT_USED
+				pool_log("die: close listen socket");
+#endif
+				close(child_inet_fd);
+			}
+			close(child_unix_fd);
+
+#ifdef NOT_USED
 			if (idle == 0)
 			{
 				pool_debug("child receives smart shutdown request but it's not in idle state");
 			}
+#endif
 			break;
 
 		case SIGINT:	/* fast shutdown */
@@ -1173,7 +1305,9 @@ static RETSIGTYPE die(int sig)
 			child_exit(0);
 			break;
 		default:
+#ifdef NOT_USED
 			pool_error("die() received unknown signal: %d", sig);
+#endif
 			break;
 	}
 }
@@ -1188,7 +1322,9 @@ static RETSIGTYPE close_idle_connection(int sig)
 	POOL_CONNECTION_POOL *p = pool_connection_pool;
 	ConnectionInfo *info;
 
+#ifdef NOT_USED
 	pool_debug("child receives close connection request");
+#endif
 
 	for (j=0;j<pool_config->max_pool;j++, p++)
 	{
@@ -1311,6 +1447,14 @@ void pool_free_startup_packet(StartupPacket *sp)
  */
 void child_exit(int code)
 {
+	if (getpid() == mypid)
+	{
+#ifdef NOT_USED
+		pool_log("child_exit: called from pgpool main. ignored.");
+#endif
+		return;
+	}
+
 	/* count down global connection counter */
 	if (accepted)
 		connection_count_down();
@@ -1324,8 +1468,14 @@ void child_exit(int code)
 			pool_close(pool_system_db_connection()->con);
 	}
 
+	if (pool_config->memory_cache_enabled && !pool_is_shmem_cache())
+	{
+		memcached_disconnect();
+	}
+
 	/* let backend know now we are exiting */
-	send_frontend_exits();
+	if (pool_connection_pool)
+		send_frontend_exits();
 
 	exit(code);
 }
@@ -1334,7 +1484,7 @@ void child_exit(int code)
  * create a persistent connection
  */
 POOL_CONNECTION_POOL_SLOT *make_persistent_db_connection(
-	char *hostname, int port, char *dbname, char *user, char *password)
+	char *hostname, int port, char *dbname, char *user, char *password, bool retry)
 {
 	POOL_CONNECTION_POOL_SLOT *cp;
 	int fd;
@@ -1363,6 +1513,7 @@ POOL_CONNECTION_POOL_SLOT *make_persistent_db_connection(
 	if (startup_packet == NULL)
 	{
 		pool_error("make_persistent_db_connection: could not allocate memory");
+		free_persisten_db_connection_memory(cp);
 		return NULL;
 	}
 	memset(startup_packet, 0, sizeof(*startup_packet));
@@ -1371,18 +1522,20 @@ POOL_CONNECTION_POOL_SLOT *make_persistent_db_connection(
 	/*
 	 * create socket
 	 */
-	if (*hostname == '\0')
+	if (*hostname == '/')
 	{
-		fd = connect_unix_domain_socket_by_port(port, pool_config->backend_socket_dir, TRUE);
+		fd = connect_unix_domain_socket_by_port(port, hostname, retry);
 	}
 	else
 	{
-		fd = connect_inet_domain_socket_by_port(hostname, port, TRUE);
+		fd = connect_inet_domain_socket_by_port(hostname, port, retry);
 	}
 
 	if (fd < 0)
 	{
 		pool_error("make_persistent_db_connection: connection to %s(%d) failed", hostname, port);
+		free_persisten_db_connection_memory(cp);
+		free(startup_packet);
 		return NULL;
 	}
 
@@ -1399,6 +1552,9 @@ POOL_CONNECTION_POOL_SLOT *make_persistent_db_connection(
 	if (len1 >= (sizeof(startup_packet->data)-len))
 	{
 		pool_error("make_persistent_db_connection: too long user name");
+		pool_close(cp->con);
+		free_persisten_db_connection_memory(cp);
+		free(startup_packet);
 		return NULL;
 	}
 
@@ -1407,6 +1563,9 @@ POOL_CONNECTION_POOL_SLOT *make_persistent_db_connection(
 	if (len1 >= (sizeof(startup_packet->data)-len))
 	{
 		pool_error("make_persistent_db_connection: too long user name");
+		pool_close(cp->con);
+		free_persisten_db_connection_memory(cp);
+		free(startup_packet);
 		return NULL;
 	}
 
@@ -1415,6 +1574,9 @@ POOL_CONNECTION_POOL_SLOT *make_persistent_db_connection(
 	if (len1 >= (sizeof(startup_packet->data)-len))
 	{
 		pool_error("make_persistent_db_connection: too long database name");
+		pool_close(cp->con);
+		free_persisten_db_connection_memory(cp);
+		free(startup_packet);
 		return NULL;
 	}
 	len += len1;
@@ -1424,6 +1586,9 @@ POOL_CONNECTION_POOL_SLOT *make_persistent_db_connection(
 	if (cp->sp == NULL)
 	{
 		pool_error("make_persistent_db_connection: could not allocate memory");
+		pool_close(cp->con);
+		free_persisten_db_connection_memory(cp);
+		free(startup_packet);
 		return NULL;
 	}
 
@@ -1435,12 +1600,16 @@ POOL_CONNECTION_POOL_SLOT *make_persistent_db_connection(
 	if (cp->sp->database == NULL)
 	{
 		pool_error("make_persistent_db_connection: could not allocate memory");
+		pool_close(cp->con);
+		free_persisten_db_connection_memory(cp);
 		return NULL;
 	}
 	cp->sp->user = strdup(user);
 	if (cp->sp->user == NULL)
 	{
 		pool_error("make_persistent_db_connection: could not allocate memory");
+		pool_close(cp->con);
+		free_persisten_db_connection_memory(cp);
 		return NULL;
 	}
 
@@ -1451,6 +1620,8 @@ POOL_CONNECTION_POOL_SLOT *make_persistent_db_connection(
 	if (status)
 	{
 		pool_error("make_persistent_db_connection: send_startup_packet failed");
+		pool_close(cp->con);
+		free_persisten_db_connection_memory(cp);
 		return NULL;
 	}
 
@@ -1460,12 +1631,37 @@ POOL_CONNECTION_POOL_SLOT *make_persistent_db_connection(
 	if (s_do_auth(cp, password))
 	{
 		pool_error("make_persistent_db_connection: s_do_auth failed");
+		pool_close(cp->con);
+		free_persisten_db_connection_memory(cp);
 		return NULL;
 	}
 
 	return cp;
 }
 
+/*
+ * Free memory of POOL_CONNECTION_POOL_SLOT.  Should only be used in
+ * make_persistent_db_connection and discard_persistent_db_connection.
+ */
+void free_persisten_db_connection_memory(POOL_CONNECTION_POOL_SLOT *cp)
+{
+	if (!cp)
+		return;
+	if (!cp->sp)
+	{
+		free(cp);
+		return;
+	}
+	if (cp->sp->startup_packet)
+		free(cp->sp->startup_packet);
+	if (cp->sp->database)
+		free(cp->sp->database);
+	if (cp->sp->user)
+		free(cp->sp->user);
+	free(cp->sp);
+	free(cp);
+}
+
 /*
  * Discard connection and memory allocated by
  * make_persistent_db_connection().
@@ -1493,15 +1689,12 @@ void discard_persistent_db_connection(POOL_CONNECTION_POOL_SLOT *cp)
 	pool_unset_nonblock(cp->con->fd);
 
 	pool_close(cp->con);
-	free(cp->sp->startup_packet);
-	free(cp->sp->database);
-	free(cp->sp->user);
-	free(cp->sp);
-	free(cp);
+	free_persisten_db_connection_memory(cp);
 }
 
 /*
- * do authentication for cp
+ * Do authentication. Assuming the only caller is
+ * *make_persistent_db_connection().
  */
 static int s_do_auth(POOL_CONNECTION_POOL_SLOT *cp, char *password)
 {
@@ -1512,6 +1705,7 @@ static int s_do_auth(POOL_CONNECTION_POOL_SLOT *cp, char *password)
 	char state;
 	char *p;
 	int pid, key;
+	bool keydata_done;
 
 	/*
 	 * read kind expecting 'R' packet (authentication response)
@@ -1643,8 +1837,12 @@ static int s_do_auth(POOL_CONNECTION_POOL_SLOT *cp, char *password)
 	}
 
 	/*
-	 * read pid etc.
+	 * Read backend key data and wait until Reay for query arriving or
+	 * error happens.
 	 */
+
+	keydata_done = false;
+
 	for (;;)
 	{
 		status = pool_read(cp->con, &kind, sizeof(kind));
@@ -1657,6 +1855,7 @@ static int s_do_auth(POOL_CONNECTION_POOL_SLOT *cp, char *password)
 		switch (kind)
 		{
 			case 'K':	/* backend key data */
+				keydata_done = true;
 				pool_debug("s_do_auth: backend key data received");
 
 				/* read message length */
@@ -1686,21 +1885,9 @@ static int s_do_auth(POOL_CONNECTION_POOL_SLOT *cp, char *password)
 					return -1;
 				}
 				cp->key = key;
+				break;
 
-				/* read kind expecting 'Z' (ready for query) */
-				status = pool_read(cp->con, &kind, sizeof(kind));
-				if (status < 0)
-				{
-					pool_error("s_do_auth: error while reading kind");
-					return -1;
-				}
-
-				if (kind != 'Z')
-				{
-					pool_error("s_do_auth: expecting Z got %c", kind);
-					return -1;
-				}
-
+			case 'Z':	/* Ready for query */
 				/* read message length */
 				status = pool_read(cp->con, &length, sizeof(length));
 				if (status < 0)
@@ -1720,16 +1907,22 @@ static int s_do_auth(POOL_CONNECTION_POOL_SLOT *cp, char *password)
 
 				pool_debug("s_do_auth: transaction state: %c", state);
 				cp->con->tstate = state;
+
+				if (!keydata_done)
+				{
+					pool_error("s_do_auth: ready for query arrived before receiving keydata");
+				}
 				return 0;
 				break;
 
 			case 'S':	/* parameter status */
-				pool_debug("s_do_auth: parameter status data received");
-
+			case 'N':	/* notice response */
+			case 'E':	/* error response */
+				/* Just throw away data */
 				status = pool_read(cp->con, &length, sizeof(length));
 				if (status < 0)
 				{
-					pool_error("s_do_auth: error while reading message length");
+					pool_error("s_do_auth: error while reading message length. kind:%c", kind);
 					return -1;
 				}
 
@@ -1742,7 +1935,7 @@ static int s_do_auth(POOL_CONNECTION_POOL_SLOT *cp, char *password)
 				break;
 
 			default:
-				pool_error("s_do_auth: unknown response \"%c\" before processing BackendKeyData",
+				pool_error("s_do_auth: unknown response \"%c\" while processing BackendKeyData",
 						   kind);
 				break;
 		}
@@ -1756,9 +1949,17 @@ static int s_do_auth(POOL_CONNECTION_POOL_SLOT *cp, char *password)
  */
 static void connection_count_up(void)
 {
+#ifdef HAVE_SIGPROCMASK
+	sigset_t oldmask;
+#else
+	int	oldmask;
+#endif
+
+	POOL_SETMASK2(&BlockSig, &oldmask);
 	pool_semaphore_lock(CONN_COUNTER_SEM);
 	Req_info->conn_counter++;
 	pool_semaphore_unlock(CONN_COUNTER_SEM);
+	POOL_SETMASK(&oldmask);
 }
 
 /*
@@ -1767,6 +1968,13 @@ static void connection_count_up(void)
  */
 static void connection_count_down(void)
 {
+#ifdef HAVE_SIGPROCMASK
+	sigset_t oldmask;
+#else
+	int	oldmask;
+#endif
+
+	POOL_SETMASK2(&BlockSig, &oldmask);
 	pool_semaphore_lock(CONN_COUNTER_SEM);
 	/*
 	 * Make sure that we do not decrement too much.  If failed to read
@@ -1779,6 +1987,7 @@ static void connection_count_down(void)
 	if (Req_info->conn_counter > 0)
 		Req_info->conn_counter--;
 	pool_semaphore_unlock(CONN_COUNTER_SEM);
+	POOL_SETMASK(&oldmask);
 }
 
 /*
@@ -1810,7 +2019,13 @@ int select_load_balancing_node(void)
 			total_weight += BACKEND_INFO(i).backend_weight;
 		}
 	}
+
+#if defined(sun) || defined(__sun)
+	r = (((double)rand())/RAND_MAX) * total_weight;
+#else
 	r = (((double)random())/RAND_MAX) * total_weight;
+#endif
+
 	total_weight = 0.0;
 	for (i=0;i<NUM_BACKENDS;i++)
 	{
@@ -1860,8 +2075,9 @@ static void init_system_db_connection(void)
 {	
 	if (pool_config->parallel_mode || pool_config->enable_query_cache)
 	{
-		system_db_connect();
-		if (PQstatus(system_db_info->pgconn) != CONNECTION_OK)
+		int nRet;
+		nRet = system_db_connect();
+		if (nRet || PQstatus(system_db_info->pgconn) != CONNECTION_OK)
 		{
 			pool_error("Could not make persistent libpq system DB connection");
 		}
@@ -1870,10 +2086,31 @@ static void init_system_db_connection(void)
 																   pool_config->system_db_port,
 																   pool_config->system_db_dbname,
 																   pool_config->system_db_user,
-																   pool_config->system_db_password);
+																   pool_config->system_db_password, false);
 		if (system_db_info->connection == NULL)
 		{
 			pool_error("Could not make persistent system DB connection");
 		}
 	}
 }
+
+/*
+ * Initialize my backend status and master node id.
+ * We copy the backend status to private area so that
+ * they are not changed while I am alive.
+ */
+void pool_initialize_private_backend_status(void)
+{
+	int i;
+
+	pool_debug("pool_initialize_private_backend_status: initialize backend status");
+
+	for (i=0;i<MAX_NUM_BACKENDS;i++)
+	{
+		private_backend_status[i] = BACKEND_INFO(i).backend_status;
+		/* my_backend_status is referred to by VALID_BACKEND macro. */
+		my_backend_status[i] = &private_backend_status[i];
+	}
+
+	my_master_node_id = REAL_MASTER_NODE_ID;
+}
